"settingsbar.py - Settings bar for the Coloromatic demo app.\n"

# ~ Type Checking (Pyright and MyPy) - Strict Mode
# ~ Linting - Ruff
# ~ Formatting - Black - max 110 characters / line

from __future__ import annotations

# Textual imports
from textual import on  # , log
from textual.app import ComposeResult
from textual.containers import Horizontal, Container, VerticalScroll
from textual.widget import Widget
from textual.widgets import Static, Input, Select, Switch, Button, TextArea

# Local imports
from textual_coloromatic import Coloromatic
from textual_coloromatic.demo.validators import (
    QualityValidator,
    FPSValidator,
    SizeValidator,
    ColorValidator,
)
from textual_coloromatic.demo.screens import ColorScreen


class SettingBox(Container):

    def __init__(
        self,
        widget: Widget,
        label: str = "",
        label_position: str = "beside",
        widget_width: int | None = None,
    ):
        """A setting box with a label and a widget. \n
        Static position can be either 'beside' or 'under'"""

        super().__init__()
        self.widget = widget
        self.label = label
        self.label_position = label_position
        self.widget_width = widget_width

    def compose(self) -> ComposeResult:

        if self.widget_width:
            self.widget.styles.width = self.widget_width

        if self.label_position == "beside":
            with Horizontal():
                yield Static(classes="setting_filler")
                if self.label:
                    yield Static(self.label, classes="setting_label")
                yield self.widget
        elif self.label_position == "under":
            with Horizontal():
                yield Static(classes="setting_filler")
                yield self.widget
            with Horizontal(classes="under_label"):
                yield Static(classes="setting_filler")
                if self.label:
                    yield Static(self.label, classes="setting_label")
            self.add_class("setting_under")


class SettingsWidget(VerticalScroll):

    justifications = [
        ("Left", "left"),
        ("Center", "center"),
        ("Right", "right"),
    ]

    animations = [
        ("gradient", "gradient"),
        ("smooth_strobe", "smooth_strobe"),
        ("fast_strobe", "fast_strobe"),
    ]

    patterns = [
        r"^[1-9][0-9]{0,2}$",  # Number between 1-999
        r"^(100|[1-9]?[0-9])%$",  # Percentage
        r"^\d*\.?\d+fr$",  # Float followed by 'fr'
    ]

    def __init__(self, coloromatic: Coloromatic, main_window: Widget):
        super().__init__()
        self.coloromatic = coloromatic
        self.main_window = main_window

    def compose(self) -> ComposeResult:

        tiling = Switch(id="tiling_switch")
        self.width_input = Input(id="width_input", validators=[SizeValidator()], max_length=5)
        self.height_input = Input(id="height_input", validators=[SizeValidator()], max_length=5)
        color_popup_button = Button("Enter Colors", id="color_list_button")
        animation_select = Select(self.animations, value="gradient", id="animation_select", allow_blank=False)
        animate_switch = Switch(False, id="animate_switch", disabled=True)
        horizontal_switch = Switch(id="horizontal_switch")
        reverse_switch = Switch(id="reverse_switch")
        gradient_quality = Input(
            id="gradient_quality",
            max_length=3,
            validators=[QualityValidator()],
        )
        animation_fps = Input(
            id="animation_fps",
            max_length=4,
            validators=[FPSValidator()],
        )
        child_switch = Switch(False, id="show_child")
        border_switch = Switch(id="border_switch", value=False)
        bg_coloromatic = Input(id="bg_coloromatic", validators=[ColorValidator(self.app.theme_variables)])

        yield Static("Settings", id="settings_title")
        yield Static("*=details in help (F1)", id="help_label")
        yield SettingBox(tiling, "Tiling*", widget_width=10)
        yield SettingBox(self.width_input, "Width*", widget_width=12)
        yield SettingBox(self.height_input, "Height*", widget_width=12)
        yield Static("Current size: 0", id="size_label")
        yield SettingBox(color_popup_button, "*")
        yield SettingBox(animation_select, "Animation Type*", widget_width=22, label_position="under")
        yield SettingBox(animate_switch, "Animate", widget_width=10)
        yield SettingBox(horizontal_switch, "Horizontal", widget_width=10)
        yield SettingBox(reverse_switch, "Reverse\nanimation", widget_width=10)
        yield SettingBox(gradient_quality, "Gradient\nquality*", widget_width=12)
        yield SettingBox(animation_fps, "Animation\nFPS*", widget_width=12)
        yield SettingBox(child_switch, "Show child\nwidget*", widget_width=10)
        yield SettingBox(border_switch, "Border", widget_width=10)
        yield SettingBox(bg_coloromatic, "ColorOmatic bg color", widget_width=22, label_position="under")

    @on(Switch.Changed, selector="#tiling_switch")
    async def tiling_switch_toggled(self, event: Switch.Changed) -> None:

        self.coloromatic.repeat = event.value
        if self.width_input.value == "":
            self.width_input.value = "1fr"
            await self.width_input.action_submit()
        if self.height_input.value == "":
            self.height_input.value = "1fr"
            await self.height_input.action_submit()

    @on(Input.Submitted, selector="#width_input")
    def width_input_set(self, event: Input.Submitted) -> None:

        if event.validation_result:
            if event.validation_result.is_valid:
                width = self.width_input.value if self.width_input.value != "" else "auto"
                height = self.height_input.value if self.height_input.value != "" else "auto"
                self.log(f"Setting container size to: ({width} x {height})")

                if width == "auto":
                    self.coloromatic.styles.width = width
                    self.log(f"Width set to: {self.coloromatic.styles.width}")
                else:
                    try:
                        self.coloromatic.styles.width = int(width)
                        self.log(f"Width set to integer: {self.coloromatic.styles.width}")
                    except ValueError:
                        self.coloromatic.styles.width = width
                        self.log(f"Width set to: {self.coloromatic.styles.width}")
                self.coloromatic.post_message(Coloromatic.Updated(self.coloromatic))
            else:
                failures = event.validation_result.failure_descriptions
                self.log(f"Invalid width: {failures}")
                self.notify(f"Invalid width: {failures}", markup=False)

    @on(Input.Submitted, selector="#height_input")
    def height_input_set(self, event: Input.Submitted) -> None:

        if event.validation_result:
            if event.validation_result.is_valid:
                width = self.width_input.value if self.width_input.value != "" else "auto"
                height = self.height_input.value if self.height_input.value != "" else "auto"
                self.log(f"Setting container size to: ({width} x {height})")

                if height == "auto":
                    self.coloromatic.styles.height = height
                    self.log(f"Height set to: {self.coloromatic.styles.height}")
                else:
                    try:
                        self.coloromatic.styles.height = int(height)
                        self.log(f"Height set to integer: {self.coloromatic.styles.height}")
                    except ValueError:
                        self.coloromatic.styles.height = height
                        self.log(f"Height set to: {self.coloromatic.styles.height}")
                self.coloromatic.post_message(Coloromatic.Updated(self.coloromatic))
            else:
                failures = event.validation_result.failure_descriptions
                self.log(f"Invalid height: {failures}")
                self.notify(f"Invalid height: {failures}", markup=False)

    def update_size_label(self, size: str) -> None:
        self.query_one("#size_label", Static).update(size)

    @on(Button.Pressed, "#color_list_button")
    async def color_list_button_pressed(self) -> None:

        await self.app.push_screen(ColorScreen())

    @on(Switch.Changed, selector="#animate_switch")
    def animate_switch_toggled(self, event: Switch.Changed) -> None:

        self.coloromatic.animated = event.value

    @on(Select.Changed, selector="#animation_select")
    def animation_selected(self, event: Select.Changed) -> None:

        self.coloromatic.set_animation_type(str(event.value))

    @on(Switch.Changed, selector="#horizontal_switch")
    def horizontal_switch_toggled(self, event: Switch.Changed) -> None:

        self.coloromatic.horizontal = event.value

    @on(Switch.Changed, selector="#reverse_switch")
    def reverse_switch_toggled(self, event: Switch.Changed) -> None:

        self.coloromatic.reverse = event.value

    @on(Input.Submitted, selector="#gradient_quality")
    def gradient_quality_set(self, event: Input.Submitted) -> None:
        """This must be a number between 1-100, or empty for auto.
        Auto mode will set the quality to the height of the widget."""

        if event.validation_result:
            if event.validation_result.is_valid:
                self.log(f"Gradient quality set to: {event.value if event.value else "auto"}")
                if event.value == "":
                    self.coloromatic.gradient_quality = "auto"
                else:
                    self.coloromatic.gradient_quality = int(event.value)
            else:
                failures = event.validation_result.failure_descriptions
                self.log(f"Invalid Gradient quality input: {failures}")
                self.notify(f"Invalid Gradient quality input: {failures}", markup=False)

    @on(Input.Submitted, selector="#animation_fps")
    def animation_fps_set(self, event: Input.Submitted) -> None:
        """This must be a number greater than 0 and maximum of 100."""

        if event.validation_result:
            if event.validation_result.is_valid:
                self.log(f"Animation speed set to: {event.value}")
                if event.value == "":
                    self.coloromatic.animation_fps = "auto"
                else:
                    self.coloromatic.animation_fps = float(event.value)
            else:
                failures = event.validation_result.failure_descriptions
                self.log(f"Invalid animation speed input: {failures}")
                self.notify(f"Invalid animation speed input: {failures}", markup=False)

    @on(Switch.Changed, selector="#show_child")
    def show_child_toggled(self, event: Switch.Changed) -> None:

        child = self.coloromatic.query_one(TextArea)
        child.display = event.value

    @on(Switch.Changed, selector="#border_switch")
    def border_switch_toggled(self, event: Switch.Changed) -> None:

        if event.value:
            self.coloromatic.add_class("bordered")
        else:
            self.coloromatic.remove_class("bordered")

    @on(Input.Submitted, selector="#bg_coloromatic")
    def bg_coloromatic_set(self, event: Input.Submitted) -> None:

        if event.validation_result:
            if event.validation_result.is_valid:
                self.log(f"Coloromatic background set to: {event.value if event.value else "auto"}")
                if event.value == "":
                    self.coloromatic.styles.background = None
                else:
                    if event.value.startswith("$"):
                        self.coloromatic.styles.background = self.app.theme_variables[event.value[1:]]
                    else:
                        self.coloromatic.styles.background = event.value
            else:
                failures = event.validation_result.failure_descriptions
                self.log(f"Invalid color input: {failures}")
                self.notify(f"Invalid color input: {failures}", markup=False)



"""Contains the demo app.
This module contains the demo application for Textual-Coloromatic.

It has its own entry script. Run with `textual-coloromatic`.
"""

# ~ Type Checking (Pyright and MyPy) - Strict Mode
# ~ Linting - Ruff
# ~ Formatting - Black - max 110 characters / line

# Python imports
from __future__ import annotations
from typing import Any, cast
from pathlib import Path
import random

# Textual imports
from textual import on  # , log
import textual.events as events
from textual.theme import Theme
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Container, ScrollableContainer
from textual.binding import Binding
from textual.widgets import Header, Footer, Static, Button, Select, Switch, TextArea, Input

# Textual library imports
from textual_slidecontainer import SlideContainer

# Local imports
from textual_coloromatic import Coloromatic
from textual_coloromatic.demo.datawidget import ActiveColors
from textual_coloromatic.demo.settingsbar import SettingsWidget
from textual_coloromatic.demo.screens import HelpScreen, CustomStringScreen
import textual_coloromatic.demo.demo_art


class BottomBar(Horizontal):

    def __init__(self, coloromatic: Coloromatic):
        super().__init__()
        self.coloromatic = coloromatic
        self.stored_custom_str: str = ""

        art_pkg_path = Path(next(iter(getattr(textual_coloromatic.demo.demo_art, "__path__"))))
        self.coloromatic.art_loader.add_directory(art_pkg_path)
        self.file_dict: dict[str, list[Path]] = self.coloromatic.art_loader.file_dict
        """String (key) is the directory name. The value is a list of path objects for each .txt file
        in that directory."""

        art_options: list[Path] = self.file_dict["demo_art"]
        # ? The pattern options are built into the ColorOmatic class:
        # Note you can also access them through the file_dict, but this is more convenient.
        pattern_options: dict[str, Path] = self.coloromatic.pattern_options

        self.art_options: list[tuple[str, Path]] = []
        for path in art_options:
            display_name = path.name.replace(path.suffix, "")
            self.art_options.append((display_name, path))

        self.pattern_options: list[tuple[str, Path]] = []
        for key, path in pattern_options.items():
            self.pattern_options.append((key, path))

    def compose(self) -> ComposeResult:

        yield Select(self.art_options, id="art_select", allow_blank=True)
        yield Button("Random", id="randomize_button")
        yield Button("Custom", id="custom_button")
        yield Switch(id="art_pattern_switch")
        yield Static("Toggle showing art or patterns", classes="bottombar_item")

    @on(Button.Pressed, "#randomize_button")
    def randomize_art(self) -> None:

        art_select = cast(Select[Path], self.query_one("#art_select", Select))
        current_selection = art_select.value
        possible_choices: list[Path] = []

        if self.query_one(Switch).value:  #         pattern mode
            for path in self.file_dict["patterns"]:
                if path != current_selection:
                    possible_choices.append(path)
        else:  #                                    art mode
            for path in self.file_dict["demo_art"]:
                if path != current_selection:
                    possible_choices.append(path)

        art_select.value = random.choice(possible_choices)

    @on(Button.Pressed, "#custom_button")
    def custom_str_button(self) -> None:

        self.app.push_screen(CustomStringScreen(self.stored_custom_str), callback=self.update_with_custom)

    def update_with_custom(self, new_str: str | None) -> None:

        if new_str:
            self.log.debug(f"{new_str = }")
            self.stored_custom_str = new_str
            self.coloromatic.text_input = new_str

    @on(Select.Changed, selector="#art_select")
    def art_changed(self, event: Select.Changed) -> None:

        if event.value == Select.BLANK:  # why is blank allowed again?
            return
        assert isinstance(event.value, Path)

        # Each Path object is a .txt file, here we just want the name without the suffix.
        # The pattern options are built into the ColorOmatic class, but I also added
        # in a demo_art directory for the sake of the demo. So this will look up
        # if the current choice is in the built-in patterns.
        choice = event.value.name.replace(event.value.suffix, "")
        if choice in self.coloromatic.pattern_options:
            self.coloromatic.set_pattern(choice)
        else:
            self.coloromatic.update_from_path(event.value)

    @on(Switch.Changed)
    def switch_changed(self, event: Switch.Changed) -> None:

        art_select = cast(Select[Path], self.query_one("#art_select", Select))
        if event.value:  # on = show patterns
            art_select.set_options(self.pattern_options)
        else:
            art_select.set_options(self.art_options)


class ColoromaticDemo(App[Any]):

    BINDINGS = [
        Binding("ctrl+b", "toggle_menu", "Expand/collapse the menu"),
        Binding("f1", "show_help", "Show help"),
    ]

    CSS_PATH = "styles.tcss"
    TITLE = "Textual-Color-O-Matic Demo"

    def compose(self) -> ComposeResult:

        yield ActiveColors()  # fancy data widget for the demo. Not part of the library.

        # assign coloromatic to variable for passing in to SettingsWidget and BottomBar
        coloromatic = Coloromatic(id="coloromatic")
        main_window = ScrollableContainer(id="main_window")

        # Note: Layout is horizontal. (top of styles.tcss)
        yield Header()
        with SlideContainer(id="menu_container", slide_direction="left", floating=False):
            yield SettingsWidget(coloromatic, main_window)
        with Container():
            with main_window:
                with coloromatic:
                    child_static = TextArea("Sample text. You can edit this text.", id="child_static")
                    child_static.display = False
                    yield child_static
            yield BottomBar(coloromatic)
        yield Footer()

    def on_mount(self, event: events.Mount) -> None:
        self.app.theme_changed_signal.subscribe(self, self._refresh_theme)  # type: ignore[unused-ignore]

    async def _refresh_theme(self, theme: Theme) -> None:
        settings_widget = self.query_one(SettingsWidget)
        bg_input = settings_widget.query_one("#bg_coloromatic", Input)
        await bg_input.action_submit()

    @on(Coloromatic.Updated)
    def coloromatic_updated(self, event: Coloromatic.Updated) -> None:

        self.log.debug(f"Coloromatic Updated: {event}")

        settings_widget = self.query_one(SettingsWidget)
        settings_widget.update_size_label(
            f"Current size: {event.widget.size.width}W x {event.widget.size.height}H"
        )

        # See if repeating mode was just set. This would be set automatically by using
        # built in patterns. If so, we need to make sure the tiling switch is set to ON.
        if event.widget.repeat:
            tiling_switch = settings_widget.query_one("#tiling_switch", Switch)
            if not tiling_switch.value:
                tiling_switch.value = True
                # tiling_switch.disabled = True
                # tiling_switch.tooltip = "Tiling is ON when using patterns."

        # If the widget is animating all the colors are removed except for one (or none),
        # it will internally stop the animation. When it does that, we need to update the
        # animate switch in the demo menu to reflect that.
        animate_switch = settings_widget.query_one("#animate_switch", Switch)
        if event.animated:
            animate_switch.value = event.animated

        active_colors = self.query_one(ActiveColors)
        if len(active_colors) <= 1:
            if animate_switch.value is True:  # safety
                animate_switch.value = False
            animate_switch.disabled = True
            animate_switch.tooltip = "Set at least 2 colors to animate."
        else:
            animate_switch.disabled = False
            animate_switch.tooltip = None

    @on(ActiveColors.Updated)
    def activecolors_updated(self, event: ActiveColors.Updated) -> None:

        active_colors = cast(ActiveColors, event.widget)
        self.log.debug(active_colors)
        self.query_one(Coloromatic).set_color_list([item[0] for item in active_colors])

    def action_toggle_menu(self) -> None:
        self.query_one(SlideContainer).toggle()

    def action_show_help(self) -> None:
        self.push_screen(HelpScreen())


def run_demo() -> None:
    """Run the demo app."""
    app = ColoromaticDemo()
    app.run()


if __name__ == "__main__":
    run_demo()



# STANDARD LIBRARY IMPORTS
from __future__ import annotations
from typing import TYPE_CHECKING
import random
if TYPE_CHECKING:
    pass

# Textual imports
from textual.app import App, ComposeResult
from textual.containers import Container
from textual.widgets import Footer, Placeholder, Static
from textual.containers import Horizontal

# Local imports
from textual_coloromatic import Coloromatic

#############
# constants #
#############
color_list = ["blue", "magenta", "yellow"]
gradient_quality = 10
fps_sides = 6
fps_tops = 15
symbol_top = "▄"
symbol_bottom = "▀"
symbol_side = "█"
#############

class TextualApp(App[None]):

    CSS = """
    Screen { align: center middle; }  
    .my_container {
        align: center middle;
        width: 50; height: 10;
    }
    .top_bottom { width: 1fr; height: 1; }
    .left_right { width: 1;   height: 1fr; }
    #pattern    { 
        align: center middle;    
        width: 50; height: 10;
        margin: 5 0;
        # border: solid $primary; 
        & > Static { width: auto; }
    }
    Placeholder { width: 1fr; height: 1fr; }      
    """

    BINDINGS = [("r", "reverse", "Reverse Animation")]

    def compose(self) -> ComposeResult:

        with Container(classes="my_container"):
            yield Coloromatic(
                symbol_top,
                repeat=True,
                gradient_quality=gradient_quality*2,
                colors=color_list,
                animate=True,
                horizontal=True,
                fps=fps_tops,
                classes="top_bottom"
            )
            with Horizontal(id="middle_row"):
                yield Coloromatic(
                    symbol_side,
                    repeat=True,
                    gradient_quality=gradient_quality,
                    colors=color_list,
                    animate=True,
                    reverse=True,
                    fps=fps_sides,
                    classes="left_right",
                )
                yield Placeholder()
                yield Coloromatic(
                    symbol_side,
                    repeat=True,
                    gradient_quality=gradient_quality,
                    colors=color_list,
                    animate=True,
                    fps=fps_sides,
                    classes="left_right",
                )
            yield Coloromatic(
                symbol_bottom,
                repeat=True,
                gradient_quality=gradient_quality*2,
                colors=color_list,
                animate=True,
                horizontal=True,
                reverse=True,
                fps=fps_tops,
                classes="top_bottom",
                id="bottom_row"               
            )  
        with Coloromatic(
            # repeat=True,
            pattern="brick2",
            colors=color_list,
            animate=True,
            id="pattern"               
        ):
            yield Static("Hello", id="pattern_text")             

        yield Footer()

    def on_mount(self):
        self.set_interval(5, self.update_text)

    def update_text(self):

        phrases = [
            "Hello, \nWorld",
            "Good stuff, \nI do say",
            "My milkshake brings \nall the boys to the yard",
            "I come from a land \ndown under",
            "Whoa, we're half way there \nTake my hand, we'll make it, I swear",
            "Never gonna give you up, \nnever gonna let you down",
        ]
        static = self.query_one("#pattern_text", Static)
        static.update(random.choice(phrases))

    def action_reverse(self):
        coloromatics = self.query(Coloromatic).results()
        for coloromatic in coloromatics:
            coloromatic.reverse = not coloromatic.reverse


TextualApp().run()



"""art_loader.py"""

# Python imports
from __future__ import annotations
from pathlib import Path


class ArtLoader:

    def __init__(self, directories: list[Path]) -> None:
        """
        Initialize the art loader with one or more art directories.

        Args:
            directories: List of directories to search for art files.
        """
        super().__init__()

        self.display = False
        self.directories: list[Path] = directories
        self.file_dict = self.load_file_dict()
        """String (key) is the directory name. The value is a list of path objects for each .txt file
        in that directory."""

    def load_file_dict(self) -> dict[str, list[Path]]:
        """Scan all file directories for .txt files. Returns a dict with each directory as
        the key and the value is a list of Path objects (one path object for each .txt file)"""
        accepted_extensions = {".txt"}

        paths_dict: dict[str, list[Path]] = {}
        for directory in self.directories:
            files: list[Path] = []
            if not directory.exists():
                continue
            for file in directory.iterdir():
                if file.is_file() and file.suffix in accepted_extensions:
                    files.append(file)
            paths_dict[directory.name] = files

        return paths_dict

    def add_directory(self, directory: Path) -> None:
        """
        Add a new directory to the art loader.

        Args:
            directory: The directory to add.
        """
        if not directory.exists() or not directory.is_dir():
            raise ValueError(f"Invalid directory: {directory}")
        self.directories.append(directory)
        self.file_dict = self.load_file_dict()

    @staticmethod
    def extract_art_at_path(path: Path) -> list[str]:
        """
        Raises:
            FileNotFoundError: If the file does not exist.
            IOError: If there is an error reading the file.
            ValueError: If the file is empty or contains only whitespace.
        """

        # the path objects will be pointing to .txt files
        # we just need to read those files and extract the contents.

        if not path.exists() or not path.is_file():
            raise FileNotFoundError(f"File not found: {path}")

        try:
            with path.open(encoding="utf-8") as file:
                art_content = file.read()
        except Exception as e:
            raise IOError(f"Error reading file {path}: {e}")
        if not art_content.strip():
            raise ValueError(f"File {path} is empty or contains only whitespace.")

        # we need to remove the header metadata from the art content
        # if the file has any, which is usually the case.
        # The header metadata is separated by a line with only dashes.

        lines: list[str] = art_content.splitlines()
        header_end_index = 0
        for index, line in enumerate(lines):
            current_index = index + 1
            if line.strip().startswith("---"):
                header_end_index = current_index
                break

        new_list: list[str] = lines[header_end_index:]
        if not art_content:
            raise ValueError(f"File {path} has no content after header.")

        return new_list
